from collections import defaultdict
import datetime
import os
from typing import List, Dict, Tuple, NamedTuple
import json
import joblib
from lkb.factory import lkb_connect, load_lkb
from sklearn.feature_extraction.text import TfidfVectorizer
from fileutils import cached_path
from nmslib.dist import FloatIndex
from lkb.lkb_base_class import Lexical_Knowledge_Base
from components.entity_linker import ContextlessLinker
from .create_index import create_tfidf_ann_index
import numpy
import nmslib
import scipy

class MentionCandidate(NamedTuple):
    """
    A data class representing a candidate entity that a mention may be linked to.
    Parameters
    ----------
    concept_id : str, required.
        The canonical concept id in the KB.
    aliases : List[str], required.
        The aliases that caused this entity to be linked.
    similarities : List[float], required.
        The cosine similarities from the mention text to the alias in tf-idf space.
    """

    concept_id: str
    term_ids: List[str]
    similarities: List[float]

class LinkerComponentsPaths(NamedTuple):
    """
    Encapsulates all the (possibly remote) paths to data for a scispacy CandidateGenerator.
    ann_index: str
        Path to the approximate nearest neighbours index.
    tfidf_vectorizer: str
        Path to the joblib serialized sklearn TfidfVectorizer.
    tfidf_vectors: str
        Path to the float-16 encoded tf-idf vectors for the entities in the KB.
    term_id_list: str
        Path to the indices mapping concepts to term_ids in the index.
    """

    ann_index_path: str
    tfidf_vectorizer_path: str
    tfidf_vectors_path: str
    term_id_list_path: str
    @classmethod
    def get_from_directory(cls,directory):
        ann_index=os.path.join(directory, "nmslib_index.bin")  # noqa
        tfidf_vectorizer=os.path.join(directory, "tfidf_vectorizer.joblib")  # noqa
        tfidf_vectors=os.path.join(directory, "tfidf_vectors_sparse.npz")  # noqa
        term_id_list=os.path.join(directory, "term_ids.json")  # noqa,
        return cls(ann_index,tfidf_vectorizer,tfidf_vectors,term_id_list)

class LinkerDefinition:
    def __init__(self,lkb_name,linker_components_path):
        self._lkb_name= lkb_name
        self._components_path = linker_components_path
    @property
    def lkb_name(self):
        return self._lkb_name
    @property
    def linker_components_path(self):
        return self._components_path
    
LINKER_PATHS = {"TRENDNET-0.1.0":LinkerDefinition("TRENDNET-0.1.0",LinkerComponentsPaths.get_from_directory("https://databrickstrenddata/lkb/linker/")),
                "TRENDNET-0.2.0":LinkerDefinition("TRENDNET-0.2.0",LinkerComponentsPaths.get_from_directory("https://databrickstrenddata/lkb/linker-0.2.0/")),
                "TRENDNET-0.2.1":LinkerDefinition("TRENDNET-0.2.1",LinkerComponentsPaths.get_from_directory("https://databrickstrenddata/lkb/linker-0.2.1/"))}


class ngramEntityLinker(ContextlessLinker):
    #Note: Can extract out the embedding logic (see vector store) and approximate KNN search.
    """
    A candidate generator for entity linking to a KnowledgeBase. Currently, two defaults are available:
     - Unified Medical Language System (UMLS).
     - Medical Subject Headings (MESH).
    To use these configured default KBs, pass the `name` parameter, either 'umls' or 'mesh'.
    It uses a sklearn.TfidfVectorizer to embed mention text into a sparse embedding of character 3-grams.
    These are then compared via cosine distance in a pre-indexed approximate nearest neighbours index of
    a subset of all entities and aliases in the KB.
    Once the K nearest neighbours have been retrieved, they are canonicalized to their KB canonical ids.
    This step is required because the index also includes entity aliases, which map to a particular canonical
    entity. This point is important for two reasons:
    1. K nearest neighbours will return a list of Y possible neighbours, where Y < K, because the entity ids
    are canonicalized.
    2. A single string may be an alias for multiple canonical entities. For example, "Jefferson County" may be an
    alias for both the canonical ids "Jefferson County, Iowa" and "Jefferson County, Texas". These are completely
    valid and important aliases to include, but it means that using the candidate generator to implement a naive
    k-nn baseline linker results in very poor performance, because there are multiple entities for some strings
    which have an exact char3-gram match, as these entities contain the same alias string. This situation results
    in multiple entities returned with a distance of 0.0, because they exactly match an alias, making a k-nn
    baseline effectively a random choice between these candidates. However, this doesn't matter if you have a
    classifier on top of the candidate generator, as is intended!
    Parameters
    ----------
    ann_index: FloatIndex
        An nmslib approximate nearest neighbours index.
    tfidf_vectorizer: TfidfVectorizer
        The vectorizer used to encode mentions.
    ann_concept_aliases_list: List[str]
        A list of strings, mapping the indices used in the ann_index to possible KB mentions.
        This is essentially used a lookup between the ann index and actual mention strings.
    kb: KnowledgeBase
        A class representing canonical concepts from the knowledge graph.
    verbose: bool
        Setting to true will print extra information about the generated candidates.
    ef_search: int
        The efs search parameter used in the index. This substantially effects runtime speed
        (higher is slower but slightly more accurate). Note that this parameter is ignored
        if a preconstructed ann_index is passed.
    name: str, optional (default = None)
        The name of the pretrained entity linker to load. Must be one of 'umls' or 'mesh'.
    """

    def __init__(
        self,
        ann_index: FloatIndex,
        tfidf_vectorizer: TfidfVectorizer,
        ann_term_ids_list: List[str],
        lkb: Lexical_Knowledge_Base,
        verbose: bool = False,
        ef_search: int = 200,
        k: int = 30,
        filter_for_definitions = False,
        no_definition_threshold = 0.95,
        threshold = 0.7
    ) -> None:
        self.ann_index = ann_index
        self.vectorizer = tfidf_vectorizer 
        self.ann_term_ids_list = ann_term_ids_list
        self.lkb = lkb
        self.verbose = verbose
        self.ef_search = ef_search
        self.k = k
        self.filter_for_definitions = filter_for_definitions
        self.no_definition_threshold = no_definition_threshold
        self.threshold = threshold

    @classmethod#TODO: Need to setup configuration files for various NLP components, including LKB's
    def load(cls,
            linker_paths,
            lkb=None,
            lkb_uri=None,
            lkb_name=None,
            lkb_dao="NEO4J",
            verbose = False,
            ef_search: int = 200,
            k: int = 30,
            filter_for_definitions = False,
            no_definition_threshold = 0.95,
            threshold = 0.7
            ):
        ann_index =cls.load_approximate_nearest_neighbours_index(cached_path(linker_paths.tfidf_vectors_path),cached_path(linker_paths.ann_index_path))
        vectorizer = joblib.load(cached_path(linker_paths.tfidf_vectorizer_path))
        ann_term_ids_list =json.load(
            open(cached_path(linker_paths.term_id_list_path))
        )
        if lkb:
            lkb = lkb
        elif lkb_uri:
            lkb = lkb_connect(lkb_uri,dao=lkb_dao)
        elif lkb_name:
            lkb = load_lkb(lkb_name,dao=lkb_dao)
        return cls(ann_index,vectorizer,ann_term_ids_list,lkb,verbose=verbose,ef_search=ef_search,k=k,filter_for_definitions=filter_for_definitions,no_definition_threshold=no_definition_threshold,threshold=threshold)
    @staticmethod
    def create_index(lkb,out_path,min_df=10):
        return create_tfidf_ann_index(out_path,lkb,min_df)
        
    @staticmethod
    def load_approximate_nearest_neighbours_index(
        tfidf_vectors_path,
        ann_index_path,
        ef_search: int = 200,
    ) -> FloatIndex:
        """
        Load an approximate nearest neighbours index from disk.
        Parameters
        ----------
        linker_paths: LinkerPaths, required.
            Contains the paths to the data required for the entity linker.
        ef_search: int, optional (default = 200)
            Controls speed performance at query time. Max value is 2000,
            but reducing to around ~100 will increase query speed by an order
            of magnitude for a small performance hit.
        """
        concept_alias_tfidfs = scipy.sparse.load_npz(
            cached_path(tfidf_vectors_path)
        ).astype(numpy.float32)
        ann_index = nmslib.init(
            method="hnsw",
            space="cosinesimil_sparse",
            data_type=nmslib.DataType.SPARSE_VECTOR,
        )
        ann_index.addDataPointBatch(concept_alias_tfidfs)
        ann_index.loadIndex(cached_path(ann_index_path))
        query_time_params = {"efSearch": ef_search}
        ann_index.setQueryTimeParams(query_time_params)

        return ann_index

    def nmslib_knn_with_zero_vectors(
        self, vectors: numpy.ndarray
    ) -> Tuple[numpy.ndarray, numpy.ndarray]:
        """
        ann_index.knnQueryBatch crashes if any of the vectors is all zeros.
        This function is a wrapper around `ann_index.knnQueryBatch` that solves this problem. It works as follows:
        - remove empty vectors from `vectors`.
        - call `ann_index.knnQueryBatch` with the non-empty vectors only. This returns `neighbors`,
        a list of list of neighbors. `len(neighbors)` equals the length of the non-empty vectors.
        - extend the list `neighbors` with `None`s in place of empty vectors.
        - return the extended list of neighbors and distances.
        """
        empty_vectors_boolean_flags = numpy.array(vectors.sum(axis=1) != 0).reshape(-1)
        empty_vectors_count = vectors.shape[0] - sum(empty_vectors_boolean_flags)
        if self.verbose:
            print(f"Number of empty vectors: {empty_vectors_count}")

        # init extended_neighbors with a list of Nones
        extended_neighbors = numpy.empty(
            (len(empty_vectors_boolean_flags),), dtype=object
        )
        extended_distances = numpy.empty(
            (len(empty_vectors_boolean_flags),), dtype=object
        )

        if vectors.shape[0] - empty_vectors_count == 0:
            return extended_neighbors, extended_distances

        # remove empty vectors before calling `ann_index.knnQueryBatch`
        vectors = vectors[empty_vectors_boolean_flags]

        # call `knnQueryBatch` to get neighbors
        original_neighbours = self.ann_index.knnQueryBatch(vectors, k=self.k)

        neighbors, distances = zip(
            *[(x[0].tolist(), x[1].tolist()) for x in original_neighbours]
        )
        neighbors = list(neighbors)
        distances = list(distances)

        # neighbors need to be converted to an np.array of objects instead of ndarray of dimensions len(vectors)xk
        # Solution: add a row to `neighbors` with any length other than k. This way, calling np.array(neighbors)
        # returns an np.array of objects
        neighbors.append([])
        distances.append([])
        # interleave `neighbors` and Nones in `extended_neighbors`
        extended_neighbors[empty_vectors_boolean_flags] = numpy.array(neighbors)[:-1]
        extended_distances[empty_vectors_boolean_flags] = numpy.array(distances)[:-1]

        return extended_neighbors, extended_distances
        
    def link_texts(self,span_texts):
        
        entity_candidates = self.get_entity_candidates(span_texts)
        results = []
        for ec_list in entity_candidates:
            predicted = []
            # concept_ids = []
            # print(ec_list)
            for entity_cand in ec_list:
                concept = self.lkb.get_concept(entity_cand.concept_id)
                score = max(entity_cand.similarities)
                if (
                    not (
                        self.filter_for_definitions
                        and concept.definition == "MISSING"
                        and self.score < self.no_definition_threshold
                    )
                ) and score > self.threshold:
                    predicted.append((concept.id, score))
            sorted_predicted = sorted(predicted, reverse=True, key=lambda x: x[1])
            # concept_ids.append()
            results.append([concept_id for concept_id, _ in sorted_predicted])
        return results
        

    def get_entity_candidates(
        self, span_texts: List[str]
    ) -> List[List[MentionCandidate]]:
        """
        Given a list of mention texts, returns a list of candidate neighbors.
        NOTE: Because we include canonical name aliases in the ann index, the list
        of candidates returned will not necessarily be of length k for each candidate,
        because we then map these to canonical ids only.
        NOTE: For a given mention, the returned candidate list might be empty, which implies that
        the tfidf vector for this mention was all zeros (i.e there were no 3 gram overlaps). This
        happens reasonably rarely, but does occasionally.
        Parameters
        ----------
        texts: List[str], required.
            The list of mention strings to generate candidates for.
        k: int, required.
            The number of ann neighbours to look up.
            Note that the number returned may differ due to aliases.
        Returns
        -------
        A list of MentionCandidate objects per mention containing KB concept_ids and aliases
        and distances which were mapped to. Note that these are lists for each concept id,
        because the index contains aliases which are canonicalized, so multiple values may map
        to the same canonical id.
        """
        if self.verbose:
            print(f"Generating candidates for {len(span_texts)} mentions")

        # tfidf vectorizer crashes on an empty array, so we return early here
        if span_texts == []:
            return []

        tfidfs = self.vectorizer.transform(span_texts)
        start_time = datetime.datetime.now()

        # `ann_index.knnQueryBatch` crashes if one of the vectors is all zeros.
        # `nmslib_knn_with_zero_vectors` is a wrapper around `ann_index.knnQueryBatch` that addresses this issue.
        batch_neighbors, batch_distances = self.nmslib_knn_with_zero_vectors(tfidfs)
        end_time = datetime.datetime.now()
        total_time = end_time - start_time
        if self.verbose:
            print(f"Finding neighbors took {total_time.total_seconds()} seconds")
        batch_mention_candidates = []
        for neighbors, distances in zip(batch_neighbors, batch_distances):
            if neighbors is None:
                neighbors = []
            if distances is None:
                distances = []

            concept_id_to_term_ids: Dict[str, List[str]] = defaultdict(list)
            concept_id_to_similarities: Dict[str, List[float]] = defaultdict(list)
            for neighbor_index, distance in zip(neighbors, distances):
                term_id = self.ann_term_ids_list[neighbor_index]
                concepts_for_mention = self.lkb.get_concepts_from_term_id(term_id)
                for concept in concepts_for_mention:
                    concept_id_to_term_ids[concept.id].append(term_id)
                    concept_id_to_similarities[concept.id].append(1.0 - distance)

            mention_candidates = [
                MentionCandidate(
                    concept_id, term_ids, concept_id_to_similarities[concept_id]
                )
                for concept_id, term_ids in concept_id_to_term_ids.items()
            ]
            mention_candidates = sorted(
                mention_candidates, key=lambda mc: mc.concept_id
            )

            batch_mention_candidates.append(mention_candidates)

        return batch_mention_candidates
    


        

